/*
 *
 * Utility to list devices
 *
 * Copyright (C) 2003 International Business Machines, Inc.
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation version 2 of the License.
 *
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License along
 *	with this program; if not, write to the Free Software Foundation, Inc.,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "sysdiag.h"

/**
 * indent: called before printing a line, it adds indent to the line up to
 *	level passed in.
 * @level: number of spaces to indent.
 */
void indent(int level)
{
	int i;
	for (i = 0; i <= level; i++)
		fprintf(stdout, " ");
}

/**
 * remove_end_newline: removes newline on the end of an attribute value
 * @value: string to remove newline from
 */
void remove_end_newline(char *value)
{
	char *p = value + (strlen(value) - 1);
	if (p != NULL && *p == '\n')
		*p = '\0';
}

/**
 * show_device_children: prints out device subdirs.
 * @sdir: print this directory's subdirectories/children.
 */
void show_device_children(struct dlist *children, int level)
{
	struct sysfs_device *child = NULL;
	
	if (children != NULL) {
		indent(level);
		fprintf(stdout, "Children:\n");
		dlist_for_each_data(children, child, struct sysfs_device) {
			indent(level+4);
			fprintf(stdout, "%s\n", child->bus_id);
		}
	}
		
}

/**
 * isbinaryvalue: checks to see if attribute is binary or not.
 * @attr: attribute to check.
 * returns 1 if binary, 0 if not.
 */
int isbinaryvalue(struct sysfs_attribute *attr)
{
	int i;

	if (attr == NULL || attr->value == NULL)
		return 0;

	for (i = 0; i < attr->len; i++)
		if (!(isascii(attr->value[i])))
			return 1;
	return 0;
}

/**
 * show_attribute_value: prints out single attribute value.
 * @attr: attricute to print.
 */
void show_attribute_value(struct sysfs_attribute *attr, int level)
{
	if (attr == NULL)
		return;
	if (attr->method & SYSFS_METHOD_SHOW) {
		if (isbinaryvalue(attr) != 0) {
			int i;
			fprintf(stdout, "\t: ");
			for (i = 0; i < attr->len; i++) {
				if (!(i %16)) {
					fprintf(stdout, "\n");
					indent(level+8);
				} else if (!(i %8))
					fprintf(stdout, " ");
				fprintf(stdout, "%02x ",
					(unsigned char)attr->value[i]);
			}
			fprintf(stdout, "\n");

		} else if (attr->value != NULL && strlen(attr->value) > 0) {
			remove_end_newline(attr->value);
			fprintf(stdout, "\t: %s\n", attr->value);
		} else
			fprintf(stdout, "\n");
	} else {
		fprintf(stdout, "\t: store method only\n");
	} 
}

/**
 * show_attribute: prints out a single attribute
 * @attr: attribute to print.
 */
void show_attribute(struct sysfs_attribute *attr, int level)
{

	if (attr == NULL) 
		return;

	if ((DIAG_OPTIONS & DIAG_SHOW_VALUE)) {
		if (!((strncmp(attr->name, "resource", 8) == 0)
				|| (strncmp(attr->name, "config", 6) == 0)
				|| (strncmp(attr->name, "pools", 5) == 0)) 
				|| (DIAG_LEVEL > 1 )) {
			if (!(attr->method & SYSFS_METHOD_SHOW))
				return;
			indent(level);
			fprintf(stdout, "%s", attr->name);
			show_attribute_value(attr, level);
		}
	} else 
		fprintf(stdout, "\n");
	
}

/**
 * show_attributes: prints out a directory's attributes.
 * @sdir: print this directory's attributes/files.
 */
void show_attributes(struct dlist *attr, int level)
{
	if (attr != NULL) {
		struct sysfs_attribute *cur = NULL;

		indent(level);
		fprintf (stdout, "Attributes:\n");
		dlist_for_each_data(attr, cur, struct sysfs_attribute) {
			show_attribute(cur, (level+4));
		}
	}
}

/**
 * show_driver_attributes: prints out driver attributes .
 * @driver: print this driver's attributes.
 */
void show_driver_attributes(struct sysfs_driver *driver, int level)
{
	if (driver != NULL) {
		struct dlist *attr = NULL;
		
		attr = sysfs_get_driver_attributes(driver);
		if (attr != NULL) {
			struct sysfs_attribute *cur = NULL;
		
			indent(level);
			fprintf (stdout, "%s Attributes:\n", driver->name);
			dlist_for_each_data(attr, cur, 
					struct sysfs_attribute) {
				show_attribute(cur, (level+4));
			}
		}
	}
}

/**
 * show_driver_devices: prints out devices under driver.
 * @driver: print devices bound to this driver.
 */
void show_driver_devices(struct sysfs_driver *driver, int level)
{
	if (driver != NULL) {
		if (driver->devices != NULL) {
			struct sysfs_device *cur = NULL;

			indent(level);
			fprintf (stdout, "Devices:\n");
			dlist_for_each_data(driver->devices, cur, 
					struct sysfs_device) {
				indent(level+4);
				fprintf (stdout, "%s\n", cur->bus_id);
			}
		}
	}
}

/**
 * show_driver: prints out driver information.
 * @driver: driver to print.
 */
void show_driver(struct sysfs_driver *driver, int level)
{
	if (driver != NULL) {
		indent(level);
		fprintf (stdout, "%s\n", driver->name);
		if (driver->devices != NULL)
			show_driver_devices(driver, (level+4));
		
		if (DIAG_OPTIONS & DIAG_SHOW_VALUE)
			show_driver_attributes(driver, (level+4));
	}
}


/**
 * show_device: prints out device information.
 * @device: device to print.
 */
void show_device(struct sysfs_device *device, int level)
{
	struct dlist *attr = NULL;
	
	if (device != NULL) {
		indent(level);
		fprintf (stdout, "%s\n", device->bus_id, device->name);
		if (device->children != NULL)
			show_device_children(device->children, (level+4));
		if (DIAG_OPTIONS & DIAG_SHOW_VALUE) {
			attr = sysfs_get_device_attributes(device);
			if (attr != NULL)
				show_attributes(attr, (level+4));
		}
		if (isalnum(device->driver_name[0])) {
			indent(level+4);
			fprintf (stdout, "Driver: %s\n", device->driver_name);
		}
	}
}

/**
 * show_root_device: prints out sys/devices device information.
 * @device: device to print.
 */
void show_root_device(struct sysfs_device *device, int level)
{
	struct dlist *attr = NULL;
	
	if (device != NULL) {
		indent(level);
		fprintf (stdout, "%s\n", device->bus_id);
		if (DIAG_OPTIONS & DIAG_SHOW_VALUE) {
			attr = sysfs_get_device_attributes(device);
			if (attr != NULL)
				show_attributes(attr, (level+4));
		}
	}
}


/**
 * show_device_tree: prints out device tree.
 * @root: root device
 */
void show_device_tree(struct sysfs_device *root, int level)
{
	if (root != NULL) {
		struct sysfs_device *cur = NULL;

		if (DEVICE_NAME[0] == 0 || (strcmp(DEVICE_NAME,
						root->bus_id) == 0)) {
			show_root_device(root, level);
		}
		if (root->children != NULL) {
			dlist_for_each_data(root->children, cur, 
					struct sysfs_device)
				show_device_tree(cur, (level+6));

		}
	}
}

/**
 *
 */

int show_devices(struct dlist *devlist, int level)
{
	struct sysfs_device *cur = NULL;
	
	dlist_for_each_data(devlist, cur, struct sysfs_device) {
		indent(level);
		fprintf(stdout, "%s\n", cur->bus_id );
		if (cur->children != NULL) {
			if (show_devices(cur->children,(level+2)) == 0) {
				return 0;
			}
		}
	}
	return -1;
}

/**
 * list_root: prints out sysfs root device tree
 * @rootname: device root to print.
 * returns 0 with success and 1 with error.
 */
int list_root(char *root_name)
{
	struct sysfs_device *dev = NULL;
	struct sysfs_root_device *root = NULL;

	if (root_name == NULL) {
		fprintf(stderr, "rootname is null\n");
		errno = EINVAL;
		return 1;
	}
	
	if ((root = sysfs_open_root_device(root_name)) == NULL) {
		fprintf(stderr, "error opening root device: %s\n", root_name);
		return -1;
	}
	
	fprintf(stdout, "\n-------------------------------------\n");
	fprintf(stdout, "Device Tree: %s\n", root_name);
	fprintf(stdout, "-------------------------------------\n");
	if (root->devices != NULL) {
		dlist_for_each_data(root->devices, dev, struct sysfs_device) {
			show_device_tree(dev, 2);
		}
	}
	
	sysfs_close_root_device(root);
	return 0;
}


/**
 *
 */
int list_bus()
{
	struct sysfs_bus *bus = NULL;
	struct sysfs_device *curdev = NULL;
	struct sysfs_driver *curdrv = NULL;

	bus = sysfs_open_bus(BUS_NAME);
	if (bus == NULL) {
		fprintf(stderr, "error opening bus: %s\n",BUS_NAME);
		return -1;
	}
	fprintf(stdout, "\n-------------------------------------\n");
	fprintf(stdout, "Bus: %s\n", BUS_NAME);
	fprintf(stdout, "-------------------------------------\n");
	
	if ((bus->devices != NULL ) && (DIAG_OPTIONS & DIAG_DEVICE)) {
		if (DEVICE_NAME[0] == 0)
			fprintf(stdout, "Devices:\n");
		dlist_for_each_data(bus->devices, curdev, struct sysfs_device) {
			if (DEVICE_NAME[0] == 0 || (strcmp(DEVICE_NAME, 
							curdev->bus_id) == 0))
				show_device(curdev, 2);
		}
	}
	if (bus->drivers != NULL && (DIAG_OPTIONS & DIAG_DRIVER)) {
		fprintf(stdout, "Drivers:\n");
		dlist_for_each_data(bus->drivers, curdrv, struct sysfs_driver) {
			if (DRIVER_NAME[0] == 0 || (strcmp(DRIVER_NAME, 
							curdrv->name)) == 0) {
					show_driver(curdrv, 2);
			}
		}
	}
	sysfs_close_bus(bus);
	return 0;

}


/**
 * show_class_device: prints out class device.
 * @dev: class device to print.
 */
void show_class_device(struct sysfs_class_device *dev, int level)
{
	struct dlist *attr = NULL;

	if (dev != NULL) {
		indent(level);
		fprintf(stdout, "%s\n", dev->name);
		if (DIAG_OPTIONS & DIAG_SHOW_VALUE) { 
			attr = sysfs_get_classdev_attributes(dev);
			if (attr != NULL)
				show_attributes(attr, (level+4));
		}
		if (dev->sysdevice != NULL && (DIAG_OPTIONS & DIAG_DEVICE))
			show_device(dev->sysdevice, (level+4));
		if (dev->driver != NULL && (DIAG_OPTIONS & DIAG_DRIVER))
			show_driver(dev->driver, (level+4));
	}
}


/**
 *
 *
 */
int list_class()
{
	struct sysfs_class *class = NULL;
	struct sysfs_class_device *cur = NULL;

	if (CLASS_NAME[0] == 0) {
		errno = EINVAL;
		return 1;
	}
	class = sysfs_open_class(CLASS_NAME);
	if (class == NULL) {
		fprintf(stderr, "Error opening class %s\n", CLASS_NAME);
		return 1;
	}
	fprintf(stdout, "\n-------------------------------------\n");
	fprintf(stdout, "Class: %s\n", CLASS_NAME);
	fprintf(stdout, "-------------------------------------\n");
	if (class->devices != NULL) {
//		if (CLASS_NAME[0] == 0)
//			fprintf(stdout, "Class Devices:\n");
		dlist_for_each_data(class->devices, cur, 
				struct sysfs_class_device) {
			if (DEVICE_NAME[0] == 0 || (strcmp(DEVICE_NAME,
							cur->name) == 0))
				show_class_device(cur, 2);
		}
	}
	sysfs_close_class(class);
	return 0;
}


/**
 *
 *
 */
/*
int show_block_queue_attributes(struct sysfs_block_device *block)
{
	struct dlist *list = NULL;
	
	list = sysfs_get_queue_attributes(block);
	if (list != NULL) {
		indent(4);
		fprintf(stdout, "queue:\n");
		show_attributes(list, 8);
	}
	list = sysfs_get_iosched_attributes(block);
	if (list != NULL) {
		indent(8);
		fprintf(stdout, "iosched:\n");
		show_attributes(list, 12);
	}
	return 0;

}
*/

/**
 * show_default_info: prints current buses, classes, and root devices
 *	supported by sysfs.
 * returns 0 with success or 1 with error.
 */
int show_default_info(void)
{
	char *data = NULL, *str = NULL;
	int j = 0,  num = 0;

	data = sysdiag_get_record(SYSDIAG_BUS_STR);
	if (data != NULL) {
		if ((num = sysdiag_db_num_of_elements(data)) != 0) {
			if (!(DIAG_OPTIONS & DIAG_SHOW_VALUE)) {
				fprintf(stdout, "supported buses:\n");
			}
			for (j = 1; j <= num; j++) {
				if ((str = sysdiag_db_get_str(data, j)) != NULL) {
					if (DIAG_OPTIONS & DIAG_SHOW_VALUE) {
						memset(BUS_NAME, 0, 
							SYSFS_NAME_LEN);
						strncpy(BUS_NAME, str, 
							strlen(str));
						list_bus();
					} else {
						indent(4);
						fprintf(stdout, "%s\n", str);
					}	
				}
			}
		}
		free(data);
		str = NULL;
		data = NULL;
	}

	data = sysdiag_get_record(SYSDIAG_CLASS_STR);
	if (data != NULL) {
		if ((num = sysdiag_db_num_of_elements(data)) != 0) {
			if (!(DIAG_OPTIONS & DIAG_SHOW_VALUE)) {
				fprintf(stdout, "supported classes:\n");
			}
			for (j = 1; j <= num; j++) {
				if ((str = sysdiag_db_get_str(data, j)) != NULL) {
					if (DIAG_OPTIONS & DIAG_SHOW_VALUE) {
						memset(CLASS_NAME, 0, 
							SYSFS_NAME_LEN);
						strncpy(CLASS_NAME, str, 
							strlen(str));
						list_class();
					} else {
						indent(4);
						fprintf(stdout, "%s\n", str);
					}	
				}
			}
		}
		free(data);
		str = NULL;
		data = NULL;
	}
	
	
	data = sysdiag_get_record(SYSFS_DEVICES_NAME);
	if (data != NULL) {
		if ((num = sysdiag_db_num_of_elements(data)) != 0) {
			if (!(DIAG_OPTIONS & DIAG_SHOW_VALUE)) {
				fprintf(stdout, "supported root devices:\n");
			}
			for (j = 1; j <= num; j++) {
				if ((str = sysdiag_db_get_str(data, j)) != NULL) {
					if (DIAG_OPTIONS & DIAG_SHOW_VALUE) {
						memset(TOPO_NAME, 0, 
							SYSFS_NAME_LEN);
						strncpy(TOPO_NAME, str, 
							strlen(str));
					list_root(TOPO_NAME);
					} else {
						indent(4);
						fprintf(stdout, "%s\n", str);
					}	
				}
					
			}
		}
		free(data);
		str = NULL;
		data = NULL;
	}
	return 0;
}


int list_main()
{
	
	if ((DIAG_OPTIONS & DIAG_DRIVER) == 0)
		DIAG_OPTIONS |= DIAG_DEVICE;
	if (BUS_NAME[0] != 0) {
		list_bus();
	}
	if (CLASS_NAME[0] != 0)
		list_class();
	if (TOPO_NAME[0] != 0)
		list_root(TOPO_NAME);
	if (BLOCK_NAME[0] == 0 && BUS_NAME[0] == 0 && CLASS_NAME[0] == 0 
			&& TOPO_NAME[0] == 0 ) {
		show_default_info();
	}
	return 0;

}

